import asyncio
import logging
from collections import defaultdict
from datetime import datetime
from time import time
from socket import gethostname
from typing import Dict, List, Set
from aiohttp.web_exceptions import HTTPNotFound, HTTPUnprocessableEntity

from org_retention.api.models import Retention, RetentionConfig, OrgConfig
from org_retention.utils.tags import Tags

_logger = logging.getLogger(__name__)

CLAIM_WAIT = 180  # secs to wait after claiming the orgs list file for update


class RetentionEngine(object):
    def __init__(self, config, persist, metrics):
        self.config = config
        self.metrics = metrics
        self.persist = persist
        self.update_started_at = 0
        self.hostname = (
            config.ecs_metadata["ContainerID"] if config.ecs_metadata else gethostname()
        )
        classes = sorted(config.getlist("retention", "classes"))
        self.defaults = {c: config.getint("retention.defaults", c) for c in classes}
        self.allowed_days = {
            c: sorted(config.getints("retention.allowed", c)) for c in classes
        }
        # ensure sane defaults
        for klass in classes:
            if not self.defaults[klass] in self.allowed_days[klass]:
                raise Exception(
                    f"Default retention for class '{klass}' not in allowed values"
                )

    async def shutdown(self, _app=None):
        pass

    async def get_org_config(self, org_key: str) -> OrgConfig:
        return OrgConfig(
            current=await self.get_all_retentions_for_org(org_key),
            defaults={k: Retention(v) for (k, v) in self.defaults.items()},
            allowed_days=self.allowed_days,
        )

    async def get_all_retentions(self) -> List[RetentionConfig]:
        return await self.persist.get_all_org_retentions()

    async def get_all_retentions_by_day_class_org(
        self,
    ) -> Dict[int, Dict[str, Set[str]]]:
        def dict_of_sets() -> Dict[str, Set[str]]:
            return defaultdict(set)

        retentions_by_day = defaultdict(dict_of_sets)

        for config in await self.persist.get_all_org_retentions():
            retentions_by_day[config.retention_days][config.retention_class].add(
                config.org
            )

        return retentions_by_day

    async def get_all_retentions_for_org(self, org_key: str) -> Dict[str, Retention]:
        retentions = {
            k: await self.get_retention(org_key, k) for k in self.defaults.keys()
        }

        return retentions

    async def get_retention(self, org_key: str, retention_class: str) -> Retention:
        if not self._is_valid_class(retention_class):
            raise HTTPNotFound(reason=f"Unknown retention class '{retention_class}'")

        days = await self.persist.get_retention_days(org_key, retention_class)
        if days is None:
            days = self.defaults[retention_class]

        return Retention(retention_days=days)

    async def set_retention(
        self, org_key: str, retention_class: str, days: int
    ) -> Retention:
        if not self._is_valid_class(retention_class):
            raise HTTPNotFound(reason=f"Unknown retention class '{retention_class}'")
        if not self._is_allowed_days(days, retention_class):
            raise HTTPUnprocessableEntity(
                reason=f"Non-allowed retention days '{days}' for class '{retention_class}'"
            )

        if days == self.defaults[retention_class]:
            await self.set_default_retention(org_key, retention_class)
        else:
            await self.persist.set_retention_days(org_key, retention_class, days)
            self.schedule_update_retention_config()

        return Retention(retention_days=days)

    async def set_default_retention(self, org_key: str, retention_class: str) -> None:
        if not self._is_valid_class(retention_class):
            raise HTTPNotFound(reason=f"Unknown retention class '{retention_class}'")

        await self.persist.clear_retention_days(org_key, retention_class)
        self.schedule_update_retention_config()

    def schedule_update_retention_config(self):
        if time() - self.update_started_at >= CLAIM_WAIT:
            asyncio.get_event_loop().create_task(self.update_retention_config())
        else:
            _logger.info(
                "Skipped updating the retention config as there is an ongoing update"
            )

    async def update_retention_config(self):
        """Update the retention config in a consistent and coordinated manner"""
        self.update_started_at = time()
        start = time()

        tags = Tags(claimer=self.hostname, claimed_at=datetime.now().isoformat())
        await self.persist.set_retention_config_file_tags(
            tags
        )  # Claim the file for update (via tags)
        await asyncio.sleep(
            CLAIM_WAIT
        )  # Give competing instances chance to make more recent claims
        current = (
            await self.persist.get_retention_config_file_tags()
        )  # Fetch the current tags of the file

        # Check if our claim is still valid (or was overwritten by more recent request of another instance)
        if not current.claimer or (
            current.claimer == tags.claimer and current.claimed_at == tags.claimed_at
        ):
            retention_config = await self._generate_retention_config()
            await self.persist.write_retention_config_file(retention_config)
            await self.metrics.histogram("update_retention_config.successful").add(
                time() - start
            )
        else:
            _logger.info(
                "Aborted update of retention config. New claimer: %s, claimed at: %s",
                current.claimer,
                current.claimed_at,
            )

    async def get_retention_class_defaults(self):
        return self.defaults

    async def get_retention_class_allowed_days(self):
        return self.allowed_days

    async def _generate_retention_config(self) -> Dict[str, Dict[str, int]]:
        """Returns a retention config in the form:
        {
            "defaults": {"retentionClass1": "retentionClass1DefaultRetentionDays", "retentionClass2": "retentionClass2DefaultRetentionDays, ... },
            "org1":     {"retentionClass1": "org1RetentionClass1RetentionDays", "retentionClass2": "org1RetentionClass2RetentionDays", ...},
            "org2":     {"retentionClass1": "org2RetentionClass1RetentionDays", "retentionClass2": "org2RetentionClass2RetentionDays", ...},
            ...
        }
        The retention config is sorted and always follows the same order: "defaults" containing sorted default retention classes on top +
        sorted orgs containing their configured retention classes (also sorted) on the bottom
        """

        orgs_retention_config = defaultdict(dict)
        orgs_retention_config["defaults"] = dict()

        for config in sorted(
            await self.persist.get_all_org_retentions(),
            key=lambda c: (c.org, c.retention_class),
        ):
            if self._is_valid_class(config.retention_class):
                orgs_retention_config[config.org][
                    config.retention_class
                ] = config.retention_days
            else:
                _logger.error(
                    "Invalid retention class: %s for org %s",
                    config.retention_class,
                    config.org,
                )

        orgs_retention_config["defaults"] = dict(self.defaults)
        return orgs_retention_config

    def _is_valid_class(self, retention_class: str):
        return retention_class in self.defaults

    def _is_allowed_days(self, days: int, retention_class: str):
        return days in self.allowed_days[retention_class]
