from aiohttp import web
import asyncio
import os

from pscr_common.log import setup_logging, clear_logging_handlers
from pscr_common.utils import MB
from pscr_common.web.auth import auth_middleware_factory, AuthClient
from pscr_common.metrics import get_reporter
from pscr_common.metadata import MetadataProvider

from org_retention.api import models
from org_retention.engine.retention_engine import RetentionEngine
from org_retention.persist.s3client import S3Client
from org_retention.utils.access_logger import FilteredAccessLogger
from org_retention.utils.config import Config
from org_retention.utils.metric_reporter_task import MetricReporterTask
from org_retention.utils.swagger import swagger_setup, build_definitions
from org_retention.web import internal_retention_routes, retention_routes, error_handler

from aio_pyformance.registry import MetricsRegistry


def build_retention_app(retention_engine):
    app = web.Application()

    table = retention_routes.build_route_table(retention_engine)
    table.setup_routes_and_auth(app)
    swagger_setup(
        app,
        "Retention API",
        api_base_url="/retention/",
        swagger_url="/doc",
        definitions=build_definitions(models),
    )

    return app


def build_internal_retention_app(retention_engine):
    app = web.Application()

    table = internal_retention_routes.build_route_table(retention_engine)
    table.setup_routes_and_auth(app)

    return app


def create_metric_reporting_task(app, metrics_registry):
    env = os.environ.get("ENV_ID", "unknown")
    env_meta = MetadataProvider().get_metadata()
    metrics_reporter = get_reporter(
        metrics_registry,
        env,
        env_meta.task_id,
        "org_retention",
        env_meta.image.split(":")[-1],
        os.getpid(),
    )
    if metrics_reporter:
        reporter_task = MetricReporterTask(metrics_reporter)
        reporter_task.register(app)


async def build_app() -> web.Application:
    config = Config()

    clear_logging_handlers()
    setup_logging(
        level=config.get("log", "level"),
        svc_version=os.environ.get("CONTAINERVERSION", "UNKNOWN"),
        env_id=os.environ.get("ENV_ID", "UNKNOWN"),
        svc_id="TODO",
    )

    metrics_registry = MetricsRegistry()

    persist = S3Client(config)
    retention_engine = RetentionEngine(config, persist, metrics_registry)

    # Convert max request size megaytes to bytes
    max_request_size_bytes = int(os.environ.get("MAX_REQUEST_SIZE_MB", 10)) * MB

    # RBAC client
    auth_client = AuthClient()

    app = web.Application(
        middlewares=[
            auth_middleware_factory(auth_client),
            error_handler.error_middleware_factory(metrics_registry),
        ],
        client_max_size=max_request_size_bytes,
        handler_args={"access_log_class": FilteredAccessLogger},
    )

    await auth_client.initialize(app)

    retention_app = build_retention_app(retention_engine)
    app.add_subapp("/retention/", retention_app)

    internal_retention_app = build_internal_retention_app(retention_engine)
    app.add_subapp("/internal/retention/", internal_retention_app)

    create_metric_reporting_task(app=app, metrics_registry=metrics_registry)

    app.on_shutdown.append(retention_engine.shutdown)

    if config.get("persist.s3.retention_config", "startup").lower() == "true":
        retention_engine.schedule_update_retention_config()

    return app


if __name__ == "__main__":
    webapp = asyncio.get_event_loop().run_until_complete(build_app())

    web.run_app(webapp, port=5004)
